#include "io.h"
#include "layout0.h"
#include "macros.h"

float
     dummy2DummyDistance=0.5,
     dummy2NodeDistance=0.75,
     node2NodeDistance=1,
     node2DummyDistance=0.75;

int dummyFactor = 2;

int compactifyLevels = 0;

int global_iterations = 0;

int phaseIterations[PHASES];
int phaseFirstUp[PHASES];
int phaseOnlyUp[PHASES];
int phaseOnlyDown[PHASES];

char* phasesNames[PHASES] = {
  "",
  "Ordering of Nodes",
  "Edge Crossings Minimization",
  "Placing of Nodes I",
  "Placing of Nodes II"
};

void(*phaseBodyDown[PHASES])(int);
void(*phaseBodyUp[PHASES])(int);

void initPhases()
{
  int i;

  for(i=0; i<PHASES; i++)
    phaseFirstUp[i]    =
    phaseOnlyUp[i]     =
    phaseOnlyDown[i]   = 0;

  phaseFirstUp[3] = 1;

  phaseIterations[1] = 1; /* ordering */
  phaseIterations[2] = 1; /* edge crossings minimization */
  phaseIterations[3] = 1; /* placing of (float) xpos */
  phaseIterations[4] = 0; /* -"- with respect to upper *and* lower levels */
}

list selectPreds(node n) { return n->preds; }
list selectSuccs(node n) { return n->succs; }
node selectFrom(edge e) { return e->from; }
node selectTo(edge e) { return e->to; }

static float selectBarycenter(node n) { return n->aux.bc; }
static float selectMedian(node n)
{
  node m;
  return (m = n->aux.median) ? m->xpos : 0;
}

/*---------------------------------------------------------------------------.
 | This is the famous barycenter routine which returns the average position  |
 | of the list of nodes (predecessors or successors) selected by the second  |
 | argument. According to the second argument which selects preds or succs   |
 | (the lists of predecessor (resp. successor) *edges*) the function given   |
 | as the third argument selects a node of an edge.                          |
 `---------------------------------------------------------------------------*/
static void barycenter1(node n)
{
  list l;
  int count;
  float bc;

  for( l=dreamPosition.get_edge_list(n), bc=0, count=0; l; l=cdr(l), count++)
    {
      node m = dreamPosition.get_node( (edge)car(l) );
      bc += m->xpos;
    }
  
  n->aux.bc = count ? bc/count : 0;
}

static void barycenter2(node n)
{
  /* don't need f and g */
  list l;
  int count;

  float bc;

  bc=0; count=0;

  for( l=n->succs; l; l=cdr(l), count++)
    {
      node m = ((edge)car(l))->to;
      bc += m->xpos;
    }

  for( l=n->preds; l; l=cdr(l), count++)
    {
      node m = ((edge)car(l))->from;
      bc += m->xpos;
    }
  
  n->aux.bc = count ? bc/count : 0;
}

static void median1() { error("median1 not implemented yet"); }
static void median2() { error("median2 not implemented yet"); }

DreamPosition dreamPosition = {
  selectBarycenter,
  barycenter1,
  barycenter1,
  barycenter2,
  selectPreds,
  selectFrom
};

static DreamPosition barycenterDreamPosition = {
  selectBarycenter,
  barycenter1,
  barycenter1,
  barycenter2,
  selectPreds,
  selectFrom
};

static DreamPosition medianDreamPosition =
  { selectMedian, median1, median1, median2 };

void useMedian() { dreamPosition = medianDreamPosition; }
void useBarycenter() { dreamPosition = barycenterDreamPosition; }

void useSinglePosition() { dreamPosition.calculate = dreamPosition.calc1; }
void useDoublePosition() { dreamPosition.calculate = dreamPosition.calc2; }

void downwardDreamPos()
{
  dreamPosition.get_edge_list = selectPreds;
  dreamPosition.get_node = selectFrom;
}

void upwardDreamPos()
{
  dreamPosition.get_edge_list = selectSuccs;
  dreamPosition.get_node = selectTo;
}

int cmpDreamPositions(const void* a, const void* b)
{
  node na = *(node*)a, nb = *(node*)b;

  if(na==nb) return 0;
  else
    {
      float af = (dreamPosition.get)(na), bf = (dreamPosition.get)(nb);
      if(af==bf)
        return na<nb ? -1 : 1;
      else
        return af < bf ? -1 : 1;
    }
}

void insertLevelIndex(int j)
{
  int max=levels[j].numberOfNodes, k; node* n=levels[j].nodes;

  for(k=0; k<max; k++)
    n[k]->xpos=k;
}

void verboseShowLevel(int l)
{
  if(verboseFlag>1)
    switch( l % 20 )
      {
        case 0: fprintf( stderr, "%d", l); break;
        case 2:
	case 4:
	case 6:
	case 8:
        case 10:
        case 12:
	case 14:
	case 16:
        case 18: fprintf( stderr, "."); break;
        default:
		 break;
      }
}

static char *dummyName = DUMMY_NODE_NAME;

node newDummyNode(int level)
{
  node n = newNode(dummyName);

#ifdef DEBUG
  if(!dummy_attr)
    error("internal in newDummyNode: you have to assign dummy_attr first!");
#endif

  n->level = level;
  n->attr = dummy_attr;

  return n;
}

int isDummy(node n) { return n->name == dummyName; }

int maxPriority = 100;

int priorityOfNode(node n)
{
  int res;
  res = list_length(n->preds);
  if(isDummy(n)) res *= dummyFactor;
  return res>=maxPriority ? maxPriority - 1 : res;
}

float nodeDistance(node a, node b)
{
  if(isDummy(a)) return isDummy(b) ? dummy2DummyDistance : dummy2NodeDistance;
  else           return isDummy(b) ? node2DummyDistance : node2NodeDistance;
}


int useStatistics = 0;

#ifndef STATISTICS

void generateStatistic() { }
void reportStatistics(int iteration, int phase) { }

#else

static float last_xpos_gm, xpos_gm;

void generateStatistic()
{
  node n;
  int i;

  last_xpos_gm = xpos_gm = 0;

  if(!nodes) return; /* just to circumvent division by zero */

  if(global_iterations==0)
    {
      for(n=nodes; n; n=cdr(n))
        n->xxpos = n->xpos;
    }
  else
  if(global_iterations==1)
    {
      for(n=nodes, i=0; n; n=cdr(n), i++)
        {
          xpos_gm +=  abs(n->xxpos - n->xpos);
	  n->xxxpos = n->xxpos;
	  n->xxpos = n->xpos;
        }
      xpos_gm /= i;
    }
  else
    {
      for(n=nodes, i=0; n; n=cdr(n), i++)
        {
          xpos_gm +=  abs(n->xxpos - n->xpos);
	  last_xpos_gm += abs(n->xxxpos - n->xxpos);
	  n->xxxpos = n->xxpos;
	  n->xxpos = n->xpos;
        }
      xpos_gm /= i; last_xpos_gm /= i;
    }
} 

void reportStatistics(int iteration, int phase)
{
  if(global_iterations>0)
    {
      fprintf(stderr,
        ")\n" VERBOSESTR "    delta to        last xpos %.3f", xpos_gm);
      if(global_iterations>1)
        fprintf(stderr,
	  "\n" VERBOSESTR "    delta to penultimate xpos %.3f", last_xpos_gm);
    }
  else
    fprintf(stderr, ")");
  if(iteration<phaseIterations[phase] - 1)
    fprintf(stderr, "\n" VERBOSESTR "  ");
}

#endif


/*---------------------------------------------------------------------------.
 | In this function we calculate the number of edgecrossings between two     |
 | nodes based on a placing on the mark field which means an integer         |
 | position.                                                                 |
 `---------------------------------------------------------------------------*/
unsigned int edge_crossings(node a, node b)
{
  unsigned int res = 0;
  list la;

#ifdef DEBUG
  if(a->level!=b->level)
    error("internal: edge_crossings between different levels");
  if(a==b)
    error("internal: edge_crossings between identical nodes");
  if(a->mark==b->mark)
    error("internal: edge_crossings between nodes with same xpos");
#endif

  for(la=a->preds; la; la=cdr(la))
    {
      int pa = ((edge) car(la))->from->mark;
      list lb;
      for(lb=b->preds; lb; lb=cdr(lb))
        {
          int pb = ((edge) car(lb))->from->mark;
	  if((pa<pb && a->mark>b->mark)||
	     (pa>pb && a->mark<b->mark))
	    res++;
	}
    }

  for(la=a->succs; la; la=cdr(la))
    {
      int pa = ((edge) car(la))->to->mark;
      list lb;
      for(lb=b->succs; lb; lb=cdr(lb))
        {
          int pb = ((edge) car(lb))->to->mark;
	  if( (pa<pb && a->mark>b->mark) ||
	      (pa>pb && a->mark<b->mark) )
	    res++;
	}
    }

  return res;
}

/*---------------------------------------------------------------------------.
 | Global number of edge crossings                                           |
 `---------------------------------------------------------------------------*/
int numberOfEdgeCrossings()
{
  int level, res = 0;

  if(verboseFlag)
    fprintf(stderr,
VERBOSESTR "  calculating number of edge crossings ...\n");
  
  for(level=0; level<numberOfLevels-1; level++)
    {
      node * n = levels[level].nodes;
      int i, num = levels[level].numberOfNodes;
      for(i=0; i<num-1; i++)
        {
	  int j;
	  for(j=i+1; j<num; j++)
	    {
	      list la;
              for(la=n[i]->succs; la; la=cdr(la))
                {
                  int pa = ((edge) car(la))->to->mark;
                  list lb;
                  for(lb=n[j]->succs; lb; lb=cdr(lb))
                    {
                      int pb = ((edge) car(lb))->to->mark;
	              if( (pa<pb && n[i]->mark>n[j]->mark) ||
	                  (pa>pb && n[i]->mark<n[j]->mark) )
	                res++;
	            }
                }
	    }
	}
    }
  return res;
}

void levelSweep(int phase)
{
  int i, level;

  if(verboseFlag && !phaseIterations[phase])
    {
      fprintf(stderr, VERBOSESTR "skipping phase %d\n", phase);
      return;
    }

  if(verboseFlag)
    fprintf( stderr, VERBOSESTR "phase %d `%s' (%d iteration%s)\n",
             phase, phasesNames[phase],
             phaseIterations[phase], phaseIterations[phase]==1 ? "" : "s");

  if(verboseFlag) fprintf( stderr, VERBOSESTR "  " );

  for(i=0; i<phaseIterations[phase]; i++, global_iterations++)
    {
      if( phaseOnlyDown[phase] ||
          ( isEven(i+phaseFirstUp[phase]) && !phaseOnlyUp[phase]) )
        {
          if(verboseFlag) fprintf( stderr, "(%dd", i+1);

          for(level=0; level<numberOfLevels; level++)
            {
	      verboseShowLevel(level);
              (phaseBodyDown[phase])(level);
            }
	}
      else
        {
          if(verboseFlag) fprintf( stderr, "(%du", i+1);

          for(level=numberOfLevels-1; level>=0; level--)
            {
	      verboseShowLevel(level);
	      (phaseBodyUp[phase])(level);
            }
	}

      if(useStatistics)
        {
	  generateStatistic();
	  reportStatistics(i,phase);
	}
      else
        {
          if(verboseFlag) fprintf( stderr, ")");
	}
    }
  if(verboseFlag)
    fprintf( stderr, "\n" VERBOSESTR "phase %d completed.\n", phase);
}

void initNodeDistances()
{
  int level;
  for(level=0; level<numberOfLevels; level++)
    {
      node* a = levels[level].nodes;
      int i, num = levels[level].numberOfNodes;

      for(i=0; i<num-1; i++)
        a[i]->distance = nodeDistance(a[i], a[i+1]);
    }
}
