#include <stdio.h>

#include "attr.h"
#include "io.h"
#include "lists.h"
#include "node.h"
#include "mem.h"

node nodes = 0;

LengthTemplate(node)

void reverseAllNodes()
{
  node l=0, n=nodes, tmp;

  while(n)
    {
      list_reverse(&n->succs);
      list_reverse(&n->preds);
      tmp = cdr(n); cdr(n)=l;
      l=n; n=tmp;
    }
  nodes=l;

  if(verboseFlag)
    {
      int len = node_length(nodes);
      fprintf(stderr, VERBOSESTR "found %d node%s\n", len, len==1 ? "" : "s" );
    }
}

CellCodeTemplate(node)

node newNode(char* s)
{
  node new = node_newCell();

  new->name = s;
  new->label = 0;

  new->succs = new->preds = 0;
  cdr(new) = nodes;
  new->attr = default_attr;
  new->level = -1;
  new->mark = 0;
  nodes = new;
  return new;
}

void disposeNode(node n)
{
  /* n -> name is owned by symbol */

  if(n->label) disposeLabel(n->label);
  freeCell(n,sizeof(_node));
}

static void printAttrRef(attribute a)
{
  if(a!=default_attr)
    {
      prC(':');
      prS(a->name);
    }
}

static void printLevel(int l)
{
  if(l!=-1)
    {
      prC('@');
      prI(l);
    }
}

void printNodeWithSuccessors(node n)
{
   list l = n->succs;
   edge e;

   /*------------------------------------------------------------------------.
    | the graph is supposed to be connected so that all nodes have to be     |
    | connected to at least one node which means that nodes with no          |
    | succesors and no defining attributes and levels have not to be written |
    | out                                                                    |
    `------------------------------------------------------------------------*/
   if( l ||
       ( n->attr != standard_attr && n->attr != default_attr ) ||
       n->level != -1
     )
     {
       prS(n->name);
       printLevel(n->level);
       printAttrRef(n->attr);
       printLabel(n->label);
       while(l)
         {
           prC('>');
           e = (edge) car(l);
           l = cdr(l);
           prS(e->to->name);
	   printAttrRef(e->attr);
         }
       prC('\n');
     }
}

void printAllNodes()
{
  node l=nodes;

  while(l)
    {
      printNodeWithSuccessors(l);
      l = cdr(l);
    }
}

edge edges = 0;

CellCodeTemplate(edge)

edge newEdge(node f, node t)
{
  edge new = edge_newCell();
  new->from = f; new->to = t;
  new->mark = 0;
  f->succs = list_cons(new,f->succs);
  t->preds = list_cons(new,t->preds);
  new->next = edges;
  edges = new;
  /*-------------------------------------------------------------------------.
   | the registratin of a new edge should always be at the front of all      |
   | edges so far. (cf generateDummies)                                      |
   `-------------------------------------------------------------------------*/
  return new;
}

void deleteEdge(edge e)
{
  list_remove(&e->to->preds, e);
  list_remove(&e->from->succs, e);
  freeCell(e, sizeof(_edge));
}

void deleteMarkedEdges()
{
  edge * previous_edge = & edges, e = edges;

  while(e)
    if(e->mark)
      {
	edge next = e -> next;
	deleteEdge(e);
	*previous_edge = next;
	e = next;
      }
    else
      {
        previous_edge = & e->next;
	e = e -> next;
      };
}

void redirectSourceOfEdge(edge e, node s, int reverse)
{
  if(reverse)
    {
      list_remove( &e->to->preds, e);
      e->to = s;
      s->preds = list_cons( e, s->preds);
    }
  else
    {
      list_remove( &e->from->succs, e);
      e->from = s;
      s->succs = list_cons( e, s->succs);
    }
}

void flipEdge(edge e)
{
  node f = e->from, t = e->to;

  list_remove(&f->succs, e);
  list_remove(&t->preds, e);

  f->preds = list_cons(e, f->preds);
  t->succs = list_cons(e, t->succs);

  e->from = t; e->to = f;
  e->attr = brotherAttr(e->attr);
}

list backedges;

void flipBackEdges()
{
  list l;

  for(l=backedges; l; l=cdr(l))
    flipEdge((edge)car(l));
}

static int numberOfSCCs()
{
  node n; int res = 0;

  for(n=nodes; n; n=cdr(n))
    if(n==n->scc) res++;

  return res;
}

#define push(i, s) list_cons((i), (s))
#define pop(s) list_removeHead((s))
#define top(s) ((int) car((s)))

static int generateSCCs_step(node n, int * dpi, list * stack)
{
  list s; int min;

  if(n->scc) return *dpi;	/* skip nodes from previous traversals */

  n->mark = n->aux.minimal_back_dpi = *dpi;

  if(!n->succs)
    {
      n -> scc = n;
      return *dpi;
    }

  min = *dpi; *dpi += 1;	/* no `++' here ??? */

  *stack = push(n, *stack);

  for(s=n->succs; s; s=cdr(s))
    {
      node to = ((edge)car(s)) -> to;
      int m;

      if(to -> mark >= 0) m = to -> aux.minimal_back_dpi;
      else m = generateSCCs_step(to, dpi, stack);

      if(m<min) min = m;
    }

  if(min==n->mark)
    {
      node m;
      do {
          m = (node) top(*stack);
          m -> scc = n;
	  m -> aux.minimal_back_dpi = min;
	  *stack = pop(*stack);
      } while(m!=n);
    }

  return min;
}

static void generateSCCs()
{
  node n;

  for(n=nodes; n; n=cdr(n))
    {
      n->mark = -1;	/* set dpi to not visited */
      n->scc = 0;
    }

  for(n=nodes; n; n=cdr(n))
    if(n->mark==-1)
      {
        int dpi = 0;
        list stack = 0;

        generateSCCs_step(n, &dpi, &stack);
      };
  
  if(verboseFlag)
    {
      fprintf(stderr,
VERBOSESTR "number of strongly connected components %d\n",
      numberOfSCCs());
    }
}

static void mark_scc_minus1(node n)
{
  list s;

  if(n->mark==-1) return;

  n->mark=-1;

  for(s=n->succs; s; s=cdr(s))
    {
      node to = ((edge)car(s))->to;
      if(to->scc==n->scc) mark_scc_minus1(to);
    }
}

static void mark_scc_scc(node n, node min)
{
  list s; node scc = n -> scc;

  if(scc == min) return;

  n->scc = min;

  for(s=n->succs; s; s=cdr(s))
    {
      node to = ((edge)car(s))->to;
      if(to->scc==scc) mark_scc_scc(to,min);
    }
}


static node get_min_repr_preds(node n)
{
  list s; node min = n;
  

  /* >= 0 serves as mark bit */

  n->mark = list_length(n->preds) + list_length(n->succs);

  for(s=n->succs; s; s=cdr(s))
    {
      node to = ((edge)car(s))->to;
      if(to->scc==n->scc && to->mark<0)
        {
          node m = get_min_repr_preds(to);
	  if(m -> mark < min -> mark)
	    min = m;
        }
    }

  return min;
}

static node get_min_level_scc(node n)
{
  list s; node min = n;

  n->mark = 1;

  for(s=n->succs; s; s=cdr(s))
    {
      node to = ((edge)car(s)) -> to;
      if(to->scc==n->scc && to->mark<0)
        {
	  node m = get_min_level_scc(to);
	  if(m->level != -1 && (min->level == -1 || m->level < min->level))
	    min = m;
	}
    }

  return min;
}

/*
 *      all nodes of sccs of n should be set to negative value
 */
static int has_preds_scc(node n)
{
  list s;
  
  if(n->mark==1) return 0;
  n->mark = 1;

  for(s=n->preds; s; s=cdr(s))
    {
      node from = ((edge)car(s))->from;
      if(from->scc!=n->scc || has_preds_scc(from)) return 1;
    }

  return 0;
}

/*---------------------------------------------------------------------------.
 | The next procedure minimizes the number of predecessors of scc            |
 | represetants. After that procedure the representant of each scc has the   |
 | minimal number of predecessors of all nodes in the scc. This means it is  |
 | probably a good choice for the highest node of the scc.                   |
 `---------------------------------------------------------------------------*/

static void minimize_scc_repr_preds()
{
  node n;

  for(n=nodes; n; n=cdr(n))
    n->mark = -2;

  for(n=nodes; n; n=cdr(n))
    if(n->mark==-2)
      {
	node min;
        mark_scc_minus1(n);
	min = get_min_level_scc(n);
	if(min -> mark==-1)
	  {
            mark_scc_minus1(n);
	    min = get_min_repr_preds(n);
	  }

	if(min!=n->scc) mark_scc_scc(n,min);
      }
}

static list findRoots()
{
  node n;
  list roots=0, l;

  for(n=nodes; n; n=cdr(n))
    n->mark = -2;

  for(n=nodes; n; n=cdr(n))
    if(n->mark==-2)
      {
        mark_scc_minus1(n);
        if(!has_preds_scc(n))
	  roots = list_cons(n->scc,roots); 
      }

  if(verboseFlag)
    {
      int d = list_length(roots);
      fprintf(stderr, VERBOSESTR "%d root%s <", d, d==1 ? "" : "s" );
      for(l=roots; l; l=cdr(l))
        fprintf(stderr, " %s", ((node)car(l))->name);
      fprintf(stderr, " >\n");
    }

  return roots;
}

void resetMarkFields()
{
  node n;
  for(n=nodes;n;n=cdr(n))
    n->mark = 0;
}

int current_index;

/*---------------------------------------------------------------------------.
 | The next procedure is actually a depth first search. It assigns to each   |
 | node a depth first index (dfi) which is defined as the number of nodes    |
 | visited before the current node in the depth first search. A backedge is  |
 | an edge which source node has an dfi lower than the dfi of its            |
 | destination node and the destination is an ancestor in the depth first    |
 | search *tree*. This means each backedge completes a cycle which means     |
 | that in this application of depth first search it has to be reversed. The |
 | other types of edges (tree, forward and cross edges edges) are not used   |
 | here.                                                                     |
 |                                                                           |
 | Before this procedure is called all mark and onThePath fields should be   |
 | set to zero, the current_index to one and the list of backedges to zero.  |
 |                                                                           |
 | Only roots may have a mark field not equal to zero.                       |
 `---------------------------------------------------------------------------*/

extern node newDummyNode(int);

static int num_self_loops;

static void _get_back_edges(node n)
{
  list s;
  edge e;

  n->mark = current_index++;
  n->aux.onThePath = 1;

  for(s=n->succs; s; s=cdr(s))
    {
      e = car(s);

#if DEBUG
      if(e->from != n) error("internal: inconsistent edge");
#endif

      if(e->to->mark)
         {
	   /* here we have to leave the recursion ! */

           if( e->to->mark <= n->mark && e->to->aux.onThePath )
             {
               /* backedge processing */

               if(e->to == n) /* self loops */
	         {
		   node m = newDummyNode(-1);
		   edge e1 = newEdge(n,m);
		   edge e2 = newEdge(m,n);
		   e2 -> attr = e -> attr;
		   correctDummyEdgesAttrs(e1,e2,0);
		   num_self_loops++;
		   e -> mark = 1;
	           if(verboseFlag>1)
		     fprintf(stderr, VERYVERBOSESTR
		             "selfloop from %s to %s\n",
       		             e->from->name, e->to->name);
		 }
	       else
	         {
                   backedges = list_cons(e, backedges);
	           if(verboseFlag>1)
		     fprintf(stderr, VERYVERBOSESTR
		             "backedge from %s to %s\n",
       		             e->from->name, e->to->name);
                 }
             }
           else
             {
               /* forwardedge processing */
             }
         }
       else
         {
           /* processing of spanning tree edge */
           _get_back_edges(e->to);
         }
    }
  n->aux.onThePath = 0;
}

list roots;

void get_back_edges()
{
  list l;
  node n;
  edge e;

  for(n=nodes; n; n=cdr(n))
    n->mark = n->aux.onThePath = 0;

  current_index = 1;
  backedges = 0;

  num_self_loops = 0;
  for(e=edges; e; e=cdr(e)) 	/* self loops will be marked */
    e->mark = 0;

  for(l=roots; l; l=cdr(l))
    _get_back_edges((node)car(l));

  if(verboseFlag)
    {
      int bl = list_length(backedges);
      fprintf(stderr, VERBOSESTR
                      "found %d backedge%s\n", bl, bl==1 ? "" : "s" );
    }

  deleteMarkedEdges();

  if(verboseFlag)
    fprintf(stderr, VERBOSESTR "found %d selfloops\n", num_self_loops);
}

void markUndirected(node n)
{
  list l;
  if(!n->mark)
    {
      n->mark = 1;

      for(l=n->succs; l; l=cdr(l))
        markUndirected( ((edge)car(l))->to );

      for(l=n->preds; l; l=cdr(l))
        markUndirected( ((edge)car(l))->from );
    }
}

/*---------------------------------------------------------------------------.
 | To test if the graph is connected we treat the graph as an undirected     |
 | graph and perform a death first search starting at *one* arbitrary node.  |
 | Than we test for each node in the graph if it has been touched.           |
 `---------------------------------------------------------------------------*/
static int _isConnected(node n)
{
  node l;

  resetMarkFields();
  markUndirected(n);
  for(l=nodes; l; l=cdr(l))
    if(!l->mark)
      return 0;
  return 1;
}

int isConnected()
{
  return nodes ? _isConnected(nodes) : 1;
}

void checkConnected()
{
  /*
  if(!isConnected())
    error("can handle only connected graphs");
  else
  if(verboseFlag)
    fprintf(stderr, VERBOSESTR "graph is connected\n");
  */

  if(verboseFlag)
    {
      if(isConnected())
        fprintf(stderr, VERBOSESTR "graph is connected\n");
      else
        fprintf(stderr, VERBOSESTR "graph is not connected\n");
     }
}

void get_roots()
{
  generateSCCs();
  minimize_scc_repr_preds();
  roots = findRoots();
}

/*---------------------------------------------------------------------------.
 | Only the ``logical'' edges are made acyclic. This determines the roots    |
 | but I do not know any other reason to do this. At least it gives you some |
 | information about your graph. All other routines should not depend on the |
 | aclycity of the graph.                                                    |
 `---------------------------------------------------------------------------*/
void makeAcyclic()
{
  get_back_edges();
  flipBackEdges();
  list_freeList(backedges);
}

int checkGraph()
{
  if(verboseFlag)
    {
      edge e;
      int c;

      for(e=edges, c=0; e; e=cdr(e), c++)
        ;
      
      fprintf(stderr, VERBOSESTR "found %d edge%s\n",
              c, c==1 ? "" : "s");
    }

  if(checkOnlyFlag)
    {
      /*---------------------------------------------------------------------.
       | printing ...                                                        |
       `---------------------------------------------------------------------*/
      ensureOutputFileIsOpen();
      printAllAttributes();
      printAllNodes();
      if(io_output) fclose(io_output);
      exit(0);
    }

  get_roots();
  checkConnected();
  makeAcyclic();		/* ATTENTION: do not assume this call */

  return 1;
}
